#include "framework.h"

mat4 TMatrix4(mat4 m) {
    return mat4(
        m.rows[0].x, m.rows[1].x, m.rows[2].x, m.rows[3].x,
        m.rows[0].y, m.rows[1].y, m.rows[2].y, m.rows[3].y,
        m.rows[0].z, m.rows[1].z, m.rows[2].z, m.rows[3].z,
        m.rows[0].w, m.rows[1].w, m.rows[2].w, m.rows[3].w
    );
}

enum MaterialType {
    ROUGH, REFLECTIVE
};

struct Material {
    vec3 ka, kd, ks;
    float shininess;
    vec3 F0;
    MaterialType type;

    Material(MaterialType t) { type = t; }
};

struct RoughMaterial : Material {
    RoughMaterial(vec3 _kd, vec3 _ks, float _shininess) : Material(ROUGH) {
        ka = _kd * M_PI;
        kd = _kd;
        ks = _ks;
        shininess = _shininess;
    }
};

vec3 operator/(vec3 num, vec3 denom) {
    return vec3(num.x / denom.x, num.y / denom.y, num.z / denom.z);
}

struct ReflectiveMaterial : Material {
    ReflectiveMaterial(vec3 n, vec3 kappa) : Material(REFLECTIVE) {
        vec3 one(1, 1, 1);
        F0 = ((n - one) * (n - one) + kappa * kappa) / ((n + one) * (n + one) + kappa * kappa);
    }
};

struct Hit {
    float t;
    vec3 position, normal;
    Material *material;

    Hit() { t = -1; }
};

struct Ray {
    vec3 start, dir;

    Ray(vec3 _start, vec3 _dir) {
        start = _start;
        dir = normalize(_dir);
    }
};

class Intersectable {
protected:
    Material *material;
public:
    virtual Hit intersect(const Ray &ray) = 0;
};

struct Sphere : public Intersectable {
    vec3 center;
    float radius;

    Sphere(const vec3 &_center, float _radius, Material *_material) {
        center = _center;
        radius = _radius;
        material = _material;
    }

    Hit intersect(const Ray &ray) {
        Hit hit;
        vec3 dist = ray.start - center;
        float a = dot(ray.dir, ray.dir);
        float b = dot(dist, ray.dir) * 2.0f;
        float c = dot(dist, dist) - radius * radius;
        float discr = b * b - 4.0f * a * c;
        if (discr < 0) return hit;
        float sqrt_discr = sqrtf(discr);
        float t1 = (-b + sqrt_discr) / 2.0f / a;    // t1 >= t2 for sure
        float t2 = (-b - sqrt_discr) / 2.0f / a;
        if (t1 <= 0) return hit;
        hit.t = (t2 > 0) ? t2 : t1;
        hit.position = ray.start + ray.dir * hit.t;
        hit.normal = (hit.position - center) * (1.0f / radius);
        hit.material = material;
        return hit;
    }
};

struct Quadric : public Intersectable {
    mat4 Q; //can be symmetric
    std::vector<Quadric *> intersectors; //if positive to an r, then there is no intersection there

    float f(vec4 r) { //r.w = 1
        return dot(r * Q, r); // = 0 for points on the surface
    }

    virtual vec3 gradf(vec4 r) { //r.w = 1
        vec4 g = r * Q * 2;
        return normalize(vec3(g.x, g.y, g.z));
    }

    void Translate(vec3 t) {
        mat4 m = TranslateMatrix(-1 * t); //inverse matrix of the transformation
        Q = m * Q * TMatrix4(m);
    }

    void Translate(float x, float y, float z) {
        Translate(vec3(x, y, z));
    }

    void Scale(vec3 s) {
        mat4 m = ScaleMatrix(vec3(1 / s.x, 1 / s.y, 1 / s.z)); //inverse matrix of the transformation
        Q = m * Q * TMatrix4(m);
    }

    void Scale(float x, float y, float z) {
        Scale(vec3(x, y, z));
    }

    void Rotate(float angle, vec3 w) {
        mat4 m = RotationMatrix(-1 * angle, w); //inverse matrix of the transformation
        Q = m * Q * TMatrix4(m);
    }

    Hit intersect(const Ray &ray) {
        Hit hit;

        vec4 S(ray.start.x, ray.start.y, ray.start.z, 1);
        vec4 D(ray.dir.x, ray.dir.y, ray.dir.z, 0);

        float a = dot(D * Q, D);
        float b = dot(D * Q, S) + dot(S * Q, D);
        float c = dot(S * Q, S);
        float discr = b * b - 4.0f * a * c;

        if (discr < 0) return hit;
        float sqrt_discr = sqrtf(discr);
        float t1 = (-b + sqrt_discr) / 2.0f / a;    // t1 >= t2 for sure
        float t2 = (-b - sqrt_discr) / 2.0f / a;

        if (t1 <= 0) return hit;

        bool t1CutOff = false;
        bool t2CutOff = false;
        for (auto cutter: intersectors) {
            vec3 t1Pos(ray.start + ray.dir * t1);
            vec4 t1HomPos(t1Pos.x, t1Pos.y, t1Pos.z, 1);
            if (cutter->f(t1HomPos) > 0) t1CutOff = true;

            vec3 t2Pos(ray.start + ray.dir * t2);
            vec4 t2HomPos(t2Pos.x, t2Pos.y, t2Pos.z, 1);
            if (cutter->f(t2HomPos) > 0) t2CutOff = true;
        }
        if (t1CutOff && t2CutOff) return hit;
        else if (t1CutOff) hit.t = t2;
        else if (t2CutOff) hit.t = t1;
        else hit.t = (t2 > 0) ? t2 : t1;

        hit.position = ray.start + ray.dir * hit.t; //+ translate;
        hit.normal = gradf(vec4(hit.position.x, hit.position.y, hit.position.z, 1));
        hit.material = material;
        return hit;
    }
};

struct Ellipsoid : public Quadric {
    float a, b, c;

    Ellipsoid(float _a, float _b, float _c, Material *_material) : a(_a), b(_b), c(_c) {
        material = _material;
        Q = mat4(
            vec4(1 / a / a, 0.0f, 0.0f, 0.0f),
            vec4(0.0f, 1 / b / b, 0.0f, 0.0f),
            vec4(0.0f, 0.0f, 1 / c / c, 0.0f),
            vec4(0.0f, 0.0f, 0.0f, -1.0f)
        );

    }
};

struct Paraboloid : public Quadric {
    float a, b;

    Paraboloid(float _a, float _b, Material *_material) : a(_a), b(_b) {
        material = _material;
        Q = mat4(
            1 / a / a, 0.0f, 0.0f, 0.0f,
            0.0f, 1 / b / b, 0.0f, 0.0f,
            0.0f, 0.0f, 0.0f, 0.0f,
            0.0f, 0.0f, 1.0f, 0.0f
        );
    }
};

struct Hyperboloid : public Quadric {
    float a, b, c;

    Hyperboloid(float _a, float _b, float _c, Material *_material) : a(_a), b(_b), c(_c) {
        material = _material;
        Q = mat4(
            1 / a / a, 0.0f, 0.0f, 0.0f,
            0.0f, 1 / b / b, 0.0f, 0.0f,
            0.0f, 0.0f, -1 / c / c, 0.0f,
            0.0f, 0.0f, 0.0f, -1.0f
        );
    }
};

struct Cylinder : public Quadric {
    float a, b;

    Cylinder(float _a, float _b, Material *_material) : a(_a), b(_b) {
        material = _material;
        Q = mat4(
            vec4(1 / a / a, 0.0f, 0.0f, 0.0f),
            vec4(0.0f, 1 / b / b, 0.0f, 0.0f),
            vec4(0.0f, 0.0f, 0.0f, 0.0f),
            vec4(0.0f, 0.0f, 0.0f, -1.0f)
        );
    }
};

struct Plane : public Quadric {
    float a;
    float b;
    float c;
    float d;
    vec3 n;

    //constructing from a point and a normal vector
    Plane(vec3 _p, vec3 _n, Material *_material) {
        material = _material;
        n = normalize(_n);
        a = n.x;
        b = n.y;
        c = n.z;
        d = -dot(_p, n);

        Q = mat4(
            vec4(0, 0, 0, 0),
            vec4(0, 0, 0, 0),
            vec4(0, 0, 0, 0),
            vec4(a, b, c, d)
        );
    }

    vec3 gradf(vec4 r) override {
        return n;
    }
};

class Camera {
    vec3 eye, lookat, right, up;
    float fov;
public:
    void set(vec3 _eye, vec3 _lookat, vec3 vup, float _fov) {
        eye = _eye;
        lookat = _lookat;
        vec3 w = eye - lookat;
        fov = _fov;
        float focus = length(w);

        right = normalize(cross(vup, w)) * focus * tanf(fov / 2);
        up = normalize(cross(w, right)) * focus * tanf(fov / 2);
    }

    Ray getRay(int X, int Y) {
        vec3 dir =
            lookat + right * (2.0f * (X + 0.5f) / windowWidth - 1) + up * (2.0f * (Y + 0.5f) / windowHeight - 1) - eye;
        return Ray(eye, dir);
    }
};

struct Light {
    vec3 direction;
    vec3 Le;

    Light(vec3 _direction, vec3 _Le) {
        direction = normalize(_direction);
        Le = _Le;
    }
};

float rnd() { return (float) rand() / RAND_MAX; }

const float epsilon = 0.0001f;
const int nLightSampling = 20;

class Scene {
    std::vector<Intersectable *> objects;
    std::vector<Light *> lights;
    Camera camera;
    vec3 La;
    vec3 L_sky;
    vec3 sun = vec3(6.0f, 6.0f, 6.0f);
    vec3 sunDir = normalize(vec3(3, -3, -2));

    std::vector<vec3> samplingPoints;
public:
    void build() {
        vec3 eye = vec3(0.0, 1.8f, -0.4f), vup = vec3(0, 0, 1), lookat = vec3(0, 0, 0);
        float fov = 80 * M_PI / 180;
        camera.set(eye, lookat, vup, fov);

        //LIGHTING
        La = vec3(0.28f, 0.28f, 0.28f);
        L_sky = vec3(0.25f, 0.25f, 0.48f);
        vec3 lightDirection(3, 2, 2), Le(1, 1, 1);
        lights.push_back(new Light(lightDirection, Le));

        //generating random points [-0.308, 0.308]x[-0.308, 0.308]
        for (int i = 0; i < nLightSampling; i++) {
            float sampleX = rnd(), sampleY = rnd();
            while (sqrt(sampleX * sampleX + sampleY * sampleY) > 1) {
                sampleX = rnd();
                sampleY = rnd();
            }
            samplingPoints.emplace_back(sampleX * 0.616f - 0.308f, sampleY * 0.616f - 0.308f, 0.95f);
        }

        //MATERIALS
        vec3 kd1(0.3f, 0.2f, 0.1f), ks(1, 1, 1);
        Material *materialRoom = new RoughMaterial(kd1, ks, 50);
        vec3 nGold(0.17f, 0.35f, 1.5f), kGold(3.1f, 2.7f, 1.9f);
        Material *materialGold = new ReflectiveMaterial(nGold, kGold);
        vec3 nSilver(0.14f, 0.16f, 0.13f), kSilver(4.1f, 2.3f, 3.1f);
        Material *materialSilver = new ReflectiveMaterial(nSilver, kSilver);
        vec3 kd2(0.2f, 0.2f, 0.4f);
        Material *materialCylinder = new RoughMaterial(kd2, ks, 50);
        vec3 kd3(0.13f, 0.4f, 0.2f);
        Material *materialVial = new RoughMaterial(kd3, ks, 50);

        //BULDING THE ROOM
        Ellipsoid *roomEllipsoid = new Ellipsoid(2.0f, 2.0f, 1, materialRoom);

        //cutting the hole at the top of the room
        Plane *topHolePlane = new Plane(vec3(0.0f, 0.0f, 0.95f), vec3(0.0f, 0.0f, 1.0f), materialCylinder);
        roomEllipsoid->intersectors.push_back(topHolePlane);

        //BUILDING THE LIGHT "TUBE"
        Quadric *lightTube = new Hyperboloid(0.6f, 0.6f, 0.3f, materialSilver);
        lightTube->Scale(1, 1, 10.0f);
        lightTube->intersectors.push_back(
            new Plane(vec3(0.0f, 0.0f, 0.95f), vec3(0.0f, 0.0f, -1.0f), materialCylinder));
        lightTube->intersectors.push_back(new Plane(vec3(0.0f, 0.0f, 4.0f), vec3(0.0f, 0.0f, 1.0f), materialCylinder));

        // BULDING THE BLUE CYLINDER
        Quadric *cylinder = new Cylinder(0.4f, 0.4f, materialCylinder);
        cylinder->intersectors.push_back(new Plane(vec3(0.0f, 0.0f, 0.3f), vec3(0.0f, 0.0f, 1.0f), materialRoom));
        cylinder->Scale(0.8, 0.8, 1.0f);
        cylinder->Translate(vec3(0.5f, -0.7f, 0.0f));

        // BUILDING THE GOLDEN PARABOLOID
        Quadric *paraboloid = new Paraboloid(0.7f, 0.85f, materialGold);
        paraboloid->Translate(-0.75f, -0.45f, 0.0f);

        //BULILDING THE GREEN VIAL
        Quadric *vial = new Hyperboloid(0.5f, 0.5f, 0.65f, materialVial);
        vial->intersectors.push_back(new Plane(vec3(0.0f, 0.0f, 0.25f), vec3(0.0f, 0.0f, 1.0f), materialRoom));
        vial->Scale(0.3f, 0.3f, 0.6f);
        vial->Scale(0.75f, 0.75f, 1.0f);
        vial->Translate(0.8f, 0.3f, 0.0f);

        // INSERTING THE OBJECTS INTO THE ROOM
        objects.push_back(roomEllipsoid);
        objects.push_back(lightTube);
        objects.push_back(cylinder);
        objects.push_back(paraboloid);
        objects.push_back(vial);
        //objects.push_back(new Sphere(vec3(0,0,0.95f), 0.616f, materialCylinder));
    }

    void render(std::vector<vec4> &image) {
        for (int Y = 0; Y < windowHeight; Y++) {
#pragma omp parallel for
            for (int X = 0; X < windowWidth; X++) {
                vec3 color = trace(camera.getRay(X, Y));
                image[Y * windowWidth + X] = vec4(color.x, color.y, color.z, 1);
            }
        }
    }

    Hit firstIntersect(Ray ray) {
        Hit bestHit;
        for (Intersectable *object : objects) {
            Hit hit = object->intersect(ray); //  hit.t < 0 if no intersection
            if (hit.t > 0 && (bestHit.t < 0 || hit.t < bestHit.t)) bestHit = hit;
        }
        if (dot(ray.dir, bestHit.normal) > 0) bestHit.normal = bestHit.normal * (-1);
        return bestHit;
    }

    bool shadowIntersect(Ray ray) {    // for directional lights
        for (Intersectable *object : objects) if (object->intersect(ray).t > 0) return true;
        return false;
    }

    vec3 trace(Ray ray, int depth = 0) {

        if (depth > 10) return L_sky + sun * pow(dot(ray.dir, sunDir), 10);

        Hit hit = firstIntersect(ray);

        if (hit.t < 0) return L_sky + sun * pow(dot(ray.dir, sunDir), 10);

        vec3 outRadiance(0, 0, 0);

        if (hit.material->type == ROUGH) {
            outRadiance = hit.material->ka * La;
            for (Light *light : lights) {
                for (vec3 samplePoint: samplingPoints) {
                    Ray shadowRay(hit.position + hit.normal * epsilon, samplePoint - hit.position);
                    if (!shadowIntersect(shadowRay)) {    // shadow computation
                        float cosTheta = fmaxf(0, dot(hit.normal, shadowRay.dir));
                        vec3 halfway = normalize(ray.dir + samplePoint);
                        float cosDelta = fmaxf(0, dot(hit.normal, halfway));
                        // {0,0,-1} is the normal vector of all the lights in the scene
                        float deltaOmega = (0.308f * 0.308f * M_PI) / nLightSampling *
                                           dot(vec3(0.0f, 0.0f, -1.0f), shadowRay.dir) /
                                           dot(samplePoint - hit.position, samplePoint - hit.position);
                        vec3 lightIn = trace(shadowRay, depth + 1);
                        vec3 brdf = hit.material->ka * (
                            hit.material->kd * cosTheta + hit.material->ks * powf(cosDelta, hit.material->shininess)
                        );
                        outRadiance = outRadiance + lightIn *  brdf * dot(shadowRay.dir, hit.normal);
                    }
                }
            }
        }

        if (hit.material->type == REFLECTIVE) {
            vec3 reflectedDir = ray.dir - hit.normal * dot(hit.normal, ray.dir) * 2.0f;
            float cosa = -dot(ray.dir, hit.normal);
            vec3 one(1, 1, 1);
            vec3 F = hit.material->F0 + (one - hit.material->F0) * pow(1 - cosa, 5);
            outRadiance = outRadiance + trace(Ray(hit.position + hit.normal * epsilon, reflectedDir), depth + 1) * F;
        }

        return outRadiance;
    }
};

GPUProgram gpuProgram; // vertex and fragment shaders
Scene scene;

// vertex shader in GLSL
const char *vertexSource = R"(
	#version 330
    precision highp float;

	layout(location = 0) in vec2 cVertexPosition;	// Attrib Array 0
	out vec2 texcoord;

	void main() {
		texcoord = (cVertexPosition + vec2(1, 1))/2;							// -1,1 to 0,1
		gl_Position = vec4(cVertexPosition.x, cVertexPosition.y, 0, 1); 		// transform to clipping space
	}
)";

// fragment shader in GLSL
const char *fragmentSource = R"(
	#version 330
    precision highp float;

	uniform sampler2D textureUnit;
	in  vec2 texcoord;			// interpolated texture coordinates
	out vec4 fragmentColor;		// output that goes to the raster memory as told by glBindFragDataLocation

	void main() {
		fragmentColor = texture(textureUnit, texcoord);
	}
)";

class FullScreenTexturedQuad {
    unsigned int vao;    // vertex array object id and texture id
    Texture texture;
public:
    FullScreenTexturedQuad(int windowWidth, int windowHeight, std::vector<vec4> &image)
        : texture(windowWidth, windowHeight, image) {
        glGenVertexArrays(1, &vao);    // create 1 vertex array object
        glBindVertexArray(vao);        // make it active

        unsigned int vbo;        // vertex buffer objects
        glGenBuffers(1, &vbo);    // Generate 1 vertex buffer objects

        // vertex coordinates: vbo0 -> Attrib Array 0 -> vertexPosition of the vertex shader
        glBindBuffer(GL_ARRAY_BUFFER, vbo); // make it active, it is an array
        float vertexCoords[] = {-1, -1, 1, -1, 1, 1, -1, 1};    // two triangles forming a quad
        glBufferData(GL_ARRAY_BUFFER, sizeof(vertexCoords), vertexCoords,
                     GL_STATIC_DRAW);       // copy to that part of the memory which is not modified
        glEnableVertexAttribArray(0);
        glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, NULL);     // stride and offset: it is tightly packed
    }

    void Draw() {
        glBindVertexArray(vao);    // make the vao and its vbos active playing the role of the data source
        gpuProgram.setUniform(texture, "textureUnit");
        glDrawArrays(GL_TRIANGLE_FAN, 0, 4);    // draw two triangles forming a quad
    }
};

FullScreenTexturedQuad *fullScreenTexturedQuad;

// Initialization, create an OpenGL context
void onInitialization() {
    glViewport(0, 0, windowWidth, windowHeight);
    scene.build();

    long timeStart = glutGet(GLUT_ELAPSED_TIME);
    std::vector<vec4> image(windowWidth * windowHeight);
    scene.render(image);
    long timeEnd = glutGet(GLUT_ELAPSED_TIME);
    printf("Rendering time: %d milliseconds\n", (timeEnd - timeStart));

    // copy image to GPU as a texture
    fullScreenTexturedQuad = new FullScreenTexturedQuad(windowWidth, windowHeight, image);

    // create program for the GPU
    gpuProgram.create(vertexSource, fragmentSource, "fragmentColor");
}

// Window has become invalid: Redraw
void onDisplay() {
    fullScreenTexturedQuad->Draw();
    glutSwapBuffers();                                    // exchange the two buffers
}

// Key of ASCII code pressed
void onKeyboard(unsigned char key, int pX, int pY) {
}

// Key of ASCII code released
void onKeyboardUp(unsigned char key, int pX, int pY) {

}

// Mouse click event
void onMouse(int button, int state, int pX, int pY) {
}

// Move mouse with key pressed
void onMouseMotion(int pX, int pY) {
}

// Idle event indicating that some time elapsed: do animation here
void onIdle() {
    //glutPostRedisplay();

}

